import pandas as pd
from pyPhases import Phase
from pyPhasesRecordloader import RecordLoader
from tqdm import tqdm

from pyPhases.Data import DataNotFound

class LoadData(Phase):
    """
    load record ids
    """

    def getMetadata(self, allRecordIds):
        recordLoader = RecordLoader.get()

        # filter non existing annotations
        metaDates = []
        metaDatesChannels = []
        for r in tqdm(allRecordIds):
            metaData = {
                "recordId": r,
                "annotationExist": recordLoader.existAnnotation(r),
                "metaDataExist": True,
            }
            metaDates.append(metaData)
            if metaData["annotationExist"]:
                try:
                    recordMetadata = recordLoader.getMetaData(r)
                except Exception as identifier:
                    metaData["metaDataExist"] = False
                    self.logError(identifier)
                    continue
                if "channels" in recordMetadata:
                    channelsMetadata = recordMetadata["channels"]
                    for cm in channelsMetadata:
                        if "recordId" not in cm:
                            cm["recordId"] = r
                        metaDatesChannels.append(cm)
                    del recordMetadata["channels"]
                metaData.update(recordMetadata)
            else:
                self.logError(f"record id {r} does not exist")

        return metaDates, metaDatesChannels

    def loadRecordIds(self, metadata, channelsMetadata):
        relevant = metadata.query("annotationExist == True")

        filterQuery = self.getConfig("dataversion.filterQuery", False)
        if filterQuery is not False:
            relevant = relevant.query(filterQuery)

        
        channelFilterQuery = self.getConfig("dataversion.channelFilterQuery", False)
        if channelFilterQuery is not False:
            filtered_channels = channelsMetadata.query(channelFilterQuery)
            
            allChannelType = self.getConfig("dataversion.allChannels", False)
            if allChannelType:
                total_type = channelsMetadata.query(f"type == '{allChannelType}'").groupby(level='recordId').size()
                valid_type = filtered_channels.groupby(level='recordId').size()
                valid_type = valid_type.reindex(total_type.index, fill_value=0)
                valid_records = total_type[total_type == valid_type].index
                relevant = relevant[relevant["recordId"].isin(valid_records)]
            else:
                # At least one channel matches (original behavior)
                relevant = relevant[relevant["recordId"].isin(filtered_channels["recordId"])]
        return relevant["recordId"].tolist()

    def generateData(self, dataName, recordId=None):
        recordLoader = RecordLoader.get()

        if dataName in ["metadata", "metadata-channels"]:
            metadataRecords = pd.DataFrame()
            metadataChannels = pd.DataFrame()
            recordLoader.setupRemoteReadOrDownload()
            if recordId is not None:
                allRecordIds = [recordId]
                mR, mC = self.getMetadata(allRecordIds)

                if dataName == "metadata":
                    return pd.DataFrame(mR)
                else:
                    return pd.DataFrame(mC)
            else:
                allRecordIds = recordLoader.getRecordList()
                if len(allRecordIds) == 0:
                    raise Exception("No records found. Check your recordLoader config and your dataversion config.")
                
                if not self.getConfig("dataIsFinal", False):
                    try:
                        metadataRecords = self.getData("metadata-tmp", pd.DataFrame, generate=False)
                        metadataChannels = self.getData("metadata-channels-tmp", pd.DataFrame, generate=False)
                        loadedRecords = metadataRecords["recordId"].to_list()
                        newRecords = [r for r in allRecordIds if r not in loadedRecords]
                        deprecatedRecords = [r for r in loadedRecords if r not in allRecordIds]

                        if len(deprecatedRecords) > 0:
                            metadataRecords = metadataRecords.drop(deprecatedRecords, errors="ignore")
                            metadataChannels = metadataChannels.drop(deprecatedRecords, errors="ignore")
                                                
                        self.logWarning(f"Not finished dataset: {len(deprecatedRecords)} deprecated records, {len(newRecords)} new records")
                        
                        allRecordIds = newRecords
                    except DataNotFound:
                        pass

            mR, mC = self.getMetadata(allRecordIds)

            # append dataframe
            if len(mR) > 0:
                metadataRecords = pd.concat([metadataRecords, pd.DataFrame(mR)]).set_index("recordId", drop=False)
            if len(mC) > 0:
                metadataChannels = pd.concat([metadataChannels, pd.DataFrame(mC)]).set_index("recordId", drop=False)

            self.project.registerData("metadata", metadataRecords, save=self.getConfig("dataIsFinal", False))
            self.project.registerData("metadata-channels", metadataChannels, save=self.getConfig("dataIsFinal", False))
            
            if not self.getConfig("dataIsFinal", False):
                self.project.registerData("metadata-tmp", metadataRecords)
                self.project.registerData("metadata-channels-tmp", metadataChannels)

        elif dataName == "allDBRecordIds":
            datasetConfig = self.getConfig("dataversion")

            if datasetConfig["recordIds"] is not None:
                recordIds = {r: [r] for r in datasetConfig["recordIds"]}
            else:
                metadata = self.getData("metadata", pd.DataFrame)
                metadataChannels = self.getData("metadata-channels", pd.DataFrame)
                recordIdsFlat = self.loadRecordIds(metadata, metadataChannels)
                recordIds = recordLoader.groupBy(datasetConfig["groupBy"], recordIdsFlat, metadata)

            if not bool(recordIds):
                raise Exception("No records found. Check your recordLoader config and your dataversion config.")

            self.project.registerData("allDBRecordIds", recordIds)

    def main(self):
        self.generateData("allDBRecordIds")



import shutil
from pathlib import Path

from pyPhases import Phase
from pyPhases.util import BatchProgress
from pyPhasesRecordloader import AnnotationNotFound, ChannelsNotPresent, Event, RecordLoader

from SleepHarmonizer.PSGEventManager import PSGEventManager
from SleepHarmonizer.recordwriter.RecordWriter import RecordWriter
from SleepHarmonizer.SignalPreprocessing import SignalPreprocessing


class Export(Phase):
    skipExisting = True
    useMultiThreading = True
    useDigitalSignals = True
    prepareEventsFor = None

    def getAnnotationsFromDf(self, df):
        return df.T.to_dict().values()

    def getLightEvents(self, lightOff, lightOn):
        lightEvents = []
        if lightOn is not None:
            lightEvents.append(Event.fromdict({"start": lightOn, "duration": 0, "name": "lightOn"}))
        if lightOff > 0:
            lightEvents.append(Event.fromdict({"start": lightOff, "duration": 0, "name": "lightOff"}))

        return lightEvents
    
    def preprocessing(self, recordSignal, preprocessingConfig):
        targetFrequency = preprocessingConfig["targetFrequency"]
        recordSignal.targetFrequency = targetFrequency

        signalPreprocessing = SignalPreprocessing.getInstance(preprocessingConfig)

        # set the target frequency for the signal to get the correct signal length
        signalPreprocessing.preprocessingSignal(recordSignal)

    def exportRecord(self, recordName):
        rw = self.recordWriter
        rl = self.recordLoader

        if self.skipExisting and rw.exist(recordName):
            self.log(f"Record {recordName} skipped because it already exists")
            return

        try:
            psgSignal, events = rl.loadRecord(recordName)
            metaData = rl.getMetaData(recordName)
            metaData["sessionId"] = rl.getSessionId(recordName)
            metaData["subjectId"] = rl.getSubjectId(recordName)
            metaData["patientCode"] = metaData["subjectId"] if "patientCod" not in metaData or metaData["patientCode"] == "" else metaData["patientCode"]
            metaData["dicom"] = rl.getDICOMMetadata(recordName)
            
            em = PSGEventManager()

            events += self.getLightEvents(rl.lightOff, rl.lightOn)
            events = em.getDataframeFromEvents(events)

            # tailor psg signal
            keepSignals = self.getConfig("export.channels", None)
            psgSignal.reduceSignals(keepSignals)

            allowEventsConfig = self.getConfig("export.annotations", None)
            allowEvents = []
            if allowEventsConfig is not None:
                for ev in allowEventsConfig:
                    if ev in em.eventGroups and ev:
                        allowEvents += em.eventGroups[ev]
                    else:
                        allowEvents.append(ev)

                events = events.query("name in @allowEvents")

            annotations = self.getAnnotationsFromDf(events)

            # signal preprocessing / harmonizing
            targetFrequency = self.preProcessingConfig["targetFrequency"]
            signalProcessor = SignalPreprocessing.getInstance(self.preProcessingConfig)
            psgSignal.targetFrequency = targetFrequency
            signalProcessor.preprocessingSignal(psgSignal)
            rw.metaData = metaData
            # write the harmonized signal into an edf
            rw.createFolderStructure()
            tmpRecordName = recordName + "-tmp"

            rw.writeSignals(tmpRecordName, psgSignal.signals, annotations, signalIsDigital=self.useDigitalSignals)
            shutil.move(rw.getFilePath(tmpRecordName), rw.getFilePath(recordName))

        except AnnotationNotFound as e:
            self.logError(f"not all required annotation exist for {recordName}")
        except ChannelsNotPresent as e:
            self.logError(f"not all required channels exist for {recordName}: {e.channels}")

    def main(self):
        self.events = None
        self.annotations = []

        self.exportPath = self.getConfig("export-path")
        if not Path(self.exportPath).exists():
            Path(self.exportPath).mkdir(parents=True, exist_ok=True)

        recordId = self.getConfig("recordId", False)
        if not recordId:
            recordIds = self.project.getData("allDBRecordIds", list)
            # flatten grouped record ids
            recordIds = [recordId for groupedRecords in recordIds.values() for recordId in groupedRecords]
        else:
            self.useMultiThreading = False
            recordIds = [recordId]

        bp = BatchProgress()
        bp.asynchronous = True
        bp.useMultiThreading = self.useMultiThreading

        self.recordLoader = RecordLoader.get()
        self.recordWriter = RecordWriter.get()
        self.recordLoader.useDigitalSignals = self.useDigitalSignals

        self.preProcessingConfig = self.getConfig("preprocessing")
        bp.start(self.exportRecord, batchList=recordIds)



import numpy as np
from pyPhases import Swappable
from pyPhasesRecordloader import RecordSignal, Signal
from pyPhasesRecordloader import SignalPreprocessing as pyPhaseSignalPreprocessing
from scipy.signal import fftconvolve, firwin, iirfilter, kaiserord, lfilter, butter, filtfilt, iirnotch, resample_poly
from scipy.interpolate import interp1d


class SignalPreprocessing(pyPhaseSignalPreprocessing, Swappable):

    def _zerophase(self, b, a, x):
        y = lfilter(b, a, x)
        y = np.flip(y)
        y = lfilter(b, a, y)
        y = np.flip(y)
        return y

    def iir(self, signal: Signal, recordSignal: RecordSignal, order, lowcut, highcut, zerophase=True):
        b, a = iirfilter(order, [lowcut, highcut], btype="bandpass", ftype="butter", fs=signal.frequency, analog=False)

        if zerophase:
            y = self._zerophase(b, a, signal.signal)
        else:
            y = lfilter(b, a, signal.signal)

        signal.signal = y

    def fftConvolution(self, signal: Signal, recordSignal: RecordSignal, kernselSeconds):
        kernel_size = int(kernselSeconds * signal.frequency) + 1

        # Compute and remove moving average with FFT convolution
        resultShape = signal.signal.shape
        center = np.zeros(resultShape)

        center = fftconvolve(signal.signal, np.ones(shape=(kernel_size,)) / kernel_size, mode="same")

        signal.signal = signal.signal - center

        # Compute and remove the rms with FFT convolution of squared signal
        scale = np.ones(resultShape)

        temp = fftconvolve(np.square(signal.signal), np.ones(shape=(kernel_size,)) / kernel_size, mode="same")

        # Deal with negative values (mathematically, it should never be negative, but fft artifacts can cause this)
        temp[temp < 0] = 0.0

        # Deal with invalid values
        invalidIndices = np.isnan(temp) | np.isinf(temp)
        temp[invalidIndices] = 0.0
        maxTemp = np.max(temp)
        temp[invalidIndices] = maxTemp

        # Finish rms calculation
        scale = np.sqrt(temp)

        # To correct records that have a zero amplitude signal
        scale[(scale == 0) | np.isinf(scale) | np.isnan(scale)] = 1.0
        signal.signal = signal.signal / scale

    def fftConvolution18m(self, signal: Signal, recordSignal: RecordSignal):
        self.fftConvolution(signal, recordSignal, 18 * 60)

    def normalizePercentage70(self, signal: Signal, recordSignal: RecordSignal):
        self.cut(signal, recordSignal, 70, 100)
        self.normalize(signal, recordSignal, 0, 1)

    def getFilterCoefficients(self, signal, tansitionWidth=15.0, cutOffHz=30.0, rippleDB=40.0):
        nyq_rate = signal.frequency / 2.0
        width = tansitionWidth / nyq_rate
        N, beta = kaiserord(rippleDB, width)
        if nyq_rate <= cutOffHz:
            cutOffHz = nyq_rate - 0.001
            self.logWarning("Cutoff frequency for FIR was adjusted to nyquist frequency.")

        # Use firwin with a Kaiser window to create a lowpass FIR filter.
        return firwin(N, cutOffHz / nyq_rate, window=("kaiser", beta))

    def antialiasingFIR(self, signal: Signal, recordSignal: RecordSignal):
        signal.signal = np.convolve(signal.signal, self.getFilterCoefficients(signal), mode="same")

    def resampleFIR(self, signal: Signal, recordSignal: RecordSignal, targetFrequency=None):
        targetFrequency = targetFrequency or recordSignal.targetFrequency
        if signal.frequency != targetFrequency:
            self.antialiasingFIR(signal, recordSignal)
            self.resample(signal, recordSignal, targetFrequency)

    def positionSHHS(self, signal: Signal, recordSignal: RecordSignal):
        uniquePositions = set(np.unique(signal.signal))
        # RIGHT, LEFT, BACK, FRONT (derived from the profusion xml, not sure if the mapping is actually correct)
        checkValues = set(uniquePositions) - {0, 1, 2, 3}
        if len(checkValues) > 0:
            # there are some records with invalid values (like shhs1-202947), we just set them to 0
            # shhs1-203716
            signal.signal[np.isin(signal.signal, list(checkValues))] = 0
            self.logError("shhs position only supports 0, 1, 2, 3 as values, conflicts: %s \n... fix in SignalPreprocessing.py" % checkValues)

        signal.signal += 10  # overwrite protection
        signal.signal[signal.signal == 10] = 5
        signal.signal[signal.signal == 11] = 3
        signal.signal[signal.signal == 12] = 2
        signal.signal[signal.signal == 13] = 4

    def positionMESA(self, signal: Signal, recordSignal: RecordSignal):
        # sourcery skip: raise-specific-error
        uniquePositions = set(np.unique(signal.signal))
        # Right, Back, Left, Front, Upright (derived from the profusion xml, not sure if the mapping is actually correct)
        checkValues = set(uniquePositions) - {0, 1, 2, 3, 4}
        if len(checkValues) > 0:
            raise Exception("domino position only supports 0, 1, 2, 3, 4 as values ... fix in SignalPreprocessing.py")

        signal.signal += 10  # overwrite protection
        signal.signal[signal.signal == 10] = 5
        signal.signal[signal.signal == 11] = 2
        signal.signal[signal.signal == 12] = 3
        signal.signal[signal.signal == 13] = 4
        signal.signal[signal.signal == 14] = 1

    def positionDomino(self, signal: Signal, recordSignal: RecordSignal):
        # sourcery skip: raise-specific-error
        uniquePositions = set(np.unique(signal.signal))
        checkValues = set(uniquePositions) - {1, 2, 3, 4, 5, 6}
        if len(checkValues) > 0:
            raise Exception("domino position only supports 1, 2, 3, 4, 5, 6 as values ... fix in SignalPreprocessing.py")

        signal.signal[signal.signal == 1] = 4
        signal.signal[signal.signal == 2] = 1
        signal.signal[signal.signal == 3] = 3
        signal.signal[signal.signal == 4] = 5
        signal.signal[signal.signal == 5] = 1
        signal.signal[signal.signal == 6] = 2

    def positionAlice(self, signal: Signal, recordSignal: RecordSignal):
        # sourcery skip: raise-specific-error
        uniquePositions = set(np.unique(signal.signal))
        checkValues = set(uniquePositions) - {0, 3, 6, 9, 12}
        if len(checkValues) > 0:
            raise Exception("alice position only supports 0, 3, 6, 9, 12 as values ... fix in SignalPreprocessing.py")

        signal.signal[signal.signal == 0] = 1
        signal.signal[signal.signal == 3] = 5
        signal.signal[signal.signal == 6] = 2
        signal.signal[signal.signal == 9] = 4
        signal.signal[signal.signal == 12] = 3

    def firEEG(self, signal: Signal, recordSignal: RecordSignal):
        self.iir(signal, recordSignal, 5, 0.5, 35, zerophase=True)

    def firEMG(self, signal: Signal, recordSignal: RecordSignal):
        self.iir(signal, recordSignal, 5, 0.5, 50, zerophase=True)



from math import ceil, floor

from pyPhasesRecordloader import AnnotationInvalid, AnnotationNotFound, Event
from pyPhasesRecordloader.recordLoaders.XMLAnnotationLoader import XMLAnnotationLoader


class AliceRMLLoader(XMLAnnotationLoader):
    smoothLightsAnnotation = 30
    stageMap = {
        "NotScored": "undefined",
        "Wake": "W",
        "NonREM1": "N1",
        "NonREM2": "N2",
        "NonREM3": "N3",
        "REM": "R",
    }

    annotationPaths = {
        "NeuroAdultAASMStaging": [
            "ScoringData",
            "StagingData",
            "UserStaging",
            "NeuroAdultAASMStaging",
        ],
        "BodyPositionState": ["BodyPositionState"],
        "Comments": ["ScoringData", "Events"],
    }

    position = {
        "Up": "Up",
        "Supine": "Supine",
        "Left": "Left",
        "Prone": "Prone",
        "Right": "Right",
    }

    graphoMap = {
        "AlphaActivity-eeg": "AlphaActivity-eeg",
        "RapidEyeMovement-eog": "RapidEyeMovement",
        "SlowEyeMovement-eog": "SlowEyeMovement",
        "Spindle-eeg": "Spindle-eeg",
    }

    eventMapArousal = {
        # "arousal_spontaneous", "arousal_plm", "", "arousal_bruxism", "arousal_noise"
        # "NasalSnore": "arousal_snore",
        "CheyneStokesRespiration": "resp_cheynestokesbreath",
        "RERA": "arousal_rera",
        "Arousal": "arousal",
    }
    eventMapMovements = {
        "LegMovement": "LegMovement",
    }
    eventMapArtefacts = {
        "ChannelFail": "ChannelFail",
    }

    eventMapApnea = {
        "ObstructiveApnea": "resp_obstructiveapnea",
        "CentralApnea": "resp_centralapnea",
        "Hypopnea": "resp_hypopnea",
        "MixedApnea": "resp_mixedapnea",
    }

    eventMapSao2 = {
        "RelativeDesaturation": "spo2_desaturation",
    }

    eventMapCardiac = {
        "HeartRateRise": "hr_rise",
    }

    def __init__(self) -> None:
        super().__init__()

        self.annotations = []
        self.metaXML = None
        self.lightOff = 0
        self.lightOn = None
        self.annotationFrequency = 1
        self.minApneaDuration = 0
        self.xmlNameSpace = "http://www.respironics.com/PatientStudy.xsd"

    def loadClassification(
        self,
        path,
        eventMap,
        eventname=lambda tag: tag.get("Type"),
        durationAttribute=None,
        startAttribute="Start",
        defaultState="ignore",
        minDuration=0,
        replaceName=None,
        filter=lambda _: True,
    ):
        tags = self.getXMLPath(self.metaXML, path)
        if tags is None:
            raise AnnotationNotFound(path)

        notused = []
        for tag in tags:
            name = eventname(tag)
            startValue = tag.get(startAttribute)
            if startValue is None:
                raise AnnotationInvalid(path + [startAttribute])
            start = int(float(startValue) * self.annotationFrequency)
            # if the name is in the eventMap it will be added to the annotations
            if name in eventMap and filter(tag):
                if replaceName:
                    eventName = replaceName(tag)
                else:
                    eventName = eventMap[name]

                if durationAttribute is not None:
                    # if there is a duration the event will be saved as as 2 events:
                    # startTime, "(eventName"
                    # endTime, "eventName)"
                    durationValue = tag.get(durationAttribute)
                    if durationValue is None:
                        raise AnnotationInvalid(path + [durationValue])

                    durationInSeconds = float(durationValue)
                    if durationInSeconds > minDuration:
                        duration = int(durationInSeconds * self.annotationFrequency)
                        self.annotations.append([start, "(" + eventName])
                        self.annotations.append([start + duration - 1, eventName + ")"])
                else:
                    # if its without a duration, it is considered a permanent state change
                    # that will persist until it is changed again
                    self.annotations.append([start, eventName])

            else:
                notused.append(name)

        return set(notused)

    def loadEvents(
        self,
        path,
        eventMap,
        eventname=lambda tag: tag.get("Type"),
        durationAttribute=None,
        startAttribute="Start",
        defaultState="ignore",
        minDuration=0,
        replaceName=None,
        filter=lambda _: True,
        extractAttributes=[],
        extractChilds=[],
    ):
        tags = self.getXMLPath(self.metaXML, path)
        if tags is None:
            raise AnnotationNotFound(path)

        events = []
        lastDefaultEvent = None
        for tag in tags:
            name = eventname(tag)
            startInSeconds = tag.get(startAttribute)

            if startInSeconds is None:
                raise AnnotationInvalid(path + [startAttribute])

            if name in eventMap and filter(tag):
                event = Event()
                event.start = float(startInSeconds)

                event.manual = not tag.get("Machine") == "true"

                data = {}
                for a in extractAttributes:
                    data[a] = tag.get(a)

                for a in extractChilds:
                    data[a] = self.getXMLPath(tag, [a]).text

                event.data = data

                if replaceName:
                    event.name = replaceName(tag)
                else:
                    event.name = eventMap[name]

                if durationAttribute is not None:
                    durationValue = tag.get(durationAttribute)
                    if durationValue is None:
                        raise AnnotationInvalid(path + [durationValue])

                    durationInSeconds = float(durationValue)

                    if durationInSeconds > minDuration:
                        event.duration = durationInSeconds
                        events.append(event)
                else:
                    if lastDefaultEvent is not None:
                        lastDefaultEvent.duration = event.start - lastDefaultEvent.start
                    events.append(event)
                    lastDefaultEvent = event

        if lastDefaultEvent is not None and self.lightOn is not None:
            lastDefaultEvent.duration = self.lightOn - lastDefaultEvent.start

        return events

    def loadAnnotation(self, rmlFile):
        self.loadXmlFile(rmlFile)
        self.annotations = []

        lightOff = self.getXMLPath(self.metaXML, ["Acquisition", "Sessions", "Session", "LightsOff"])
        self.lightOff = 0 if lightOff is None or lightOff.text is None else int(lightOff.text)

        lightOn = self.getXMLPath(self.metaXML, ["Acquisition", "Sessions", "Session", "LightsOn"])
        self.lightOn = None if lightOn is None or lightOn.text is None else int(lightOn.text)

        duration = self.getXMLPath(self.metaXML, ["Acquisition", "Sessions", "Session", "Duration"])
        self.recordDuration = None if duration is None or duration.text is None else int(duration.text)

        # the next step is required to overwrite the edf record duration since the last part is not relevant
        # check for min duration see  acq_522138949 where duration = lightOff
        if self.lightOn is None and self.recordDuration is not None and self.recordDuration > self.lightOff:
            self.lightOn = self.recordDuration

        if self.smoothLightsAnnotation > 0:
            self.lightOff = int(self.smoothLightsAnnotation * ceil(self.lightOff / self.smoothLightsAnnotation))
            if self.lightOn is not None:
                self.lightOn = int(self.smoothLightsAnnotation * floor(self.lightOn / self.smoothLightsAnnotation))

        channelNames = {
            c.get("EdfSignal"): self.getXMLPath(c, ["Label"]).text
            for c in self.getXMLPath(self.metaXML, ["ChannelConfig", "Channels"])
        }

        allEvents = []

        allEvents += self.loadEvents(
            ["ScoringData", "StagingData", "UserStaging", "NeuroAdultAASMStaging"],
            self.stageMap,
        )

        allEvents += self.loadEvents(
            ["BodyPositionState"],
            self.position,
            eventname=lambda tag: tag.get("Position"),
        )

        allEvents += self.loadEvents(
            ["ScoringData", "Events"],
            self.eventMapArousal,
            eventname=lambda tag: self.getXMLPath(tag, ["Comment"]).text if tag.get("Type") == "Comment" else tag.get("Type"),
            durationAttribute="Duration",
        )

        def getLegPosition(tag):
            channelIndex = self.getXMLPath(tag, ["LegMovement"]).get("EdfSignal")
            if channelIndex != "-1":  # channel can be "-1" see acq_552406419
                if channelIndex not in channelNames:
                    self.logError("The Channel for Legmovment does not exist!")
                    return "Left"
                channelName = channelNames[channelIndex]
                if channelName not in ["Bein li", "BeinLi", "BeinRe", "Bein re", "BEIN.re", "BEIN.li", "Bein_re", "Bein_li"]:
                    self.logWarning("Leg ChannelName %s not known" % (channelName))
                return "Left" if channelName in ["Bein li", "BeinLi", "BEIN.li", "Bein_li"] else "Right"
            else:
                return "Left"

        if "forceLegChannel" in self.classificationConfig:
            edfChannel = self.classificationConfig["forceLegChannel"]
            edfChannelId = self.channelMap[edfChannel]
            legMovements = self.loadEvents(
                ["ScoringData", "Events"],
                self.eventMapMovements,
                eventname=lambda tag: tag.get("Type"),
                durationAttribute="Duration",
                filter=lambda tag: self.getXMLPath(tag, ["LegMovement"]).get("EdfSignal") == str(edfChannelId),
            )
        else:
            legMovements = self.loadEvents(
                ["ScoringData", "Events"],
                self.eventMapMovements,
                eventname=lambda tag: tag.get("Type"),
                replaceName=lambda tag: "LegMovement-%s" % (getLegPosition(tag)),
                durationAttribute="Duration",
            )

        # allEvents.append(legMovements)
        allEvents += legMovements

        allEvents += self.loadEvents(
            ["ScoringData", "Events"],
            self.eventMapApnea,
            eventname=lambda tag: tag.get("Type"),
            durationAttribute="Duration",
            minDuration=self.minApneaDuration,
            extractAttributes=["Machine"],
        )

        allEvents += self.loadEvents(
            ["ScoringData", "Events"],
            self.eventMapArtefacts,
            eventname=lambda tag: tag.get("Type"),
            replaceName=lambda tag: "fail-channel%s" % (self.getXMLPath(tag, ["ChannelFail"]).get("EdfSignal")),
            durationAttribute="Duration",
        )

        allEvents += self.loadEvents(
            ["ScoringData", "Events"],
            self.eventMapSao2,
            eventname=lambda tag: tag.get("Type"),
            durationAttribute="Duration",
            extractChilds=["O2Before", "O2Min"],
        )

        allEvents += self.loadEvents(
            ["ScoringData", "Events"],
            self.eventMapCardiac,
            eventname=lambda tag: tag.get("Type"),
            durationAttribute="Duration",
        )

        # TODO: check for multi segments/sessions
        sessions = self.getXMLPath(self.metaXML, ["Acquisition", "Sessions"])
        segments = self.getXMLPath(self.metaXML, ["Acquisition", "Sessions", "Session", "Segment"])
        if sessions is not None and len(sessions) > 1:
            self.logError("unexpected")

        if segments is not None and len(segments) > 1:
            self.logError("unexpected")

        return allEvents

    def getMetaData(self, rmlFile):
        self.loadXmlFile(rmlFile)
        metadata = {}
        comment = self.getXMLPath(self.metaXML, ["ChannelConfig", "ConfigName"])
        metadata["comment"] = None if comment is None else comment.text

        lightOff = self.getXMLPath(self.metaXML, ["Acquisition", "Sessions", "Session", "LightsOff"])
        metadata["lightOff"] = 0 if lightOff is None or lightOff.text is None else int(lightOff.text)

        lightOn = self.getXMLPath(self.metaXML, ["Acquisition", "Sessions", "Session", "LightsOn"])
        metadata["lightOn"] = None if lightOn is None or lightOn.text is None else int(lightOn.text)
        
        return metadata




from phases.commands.test import Test as PhasesTest
from pyPhases.test import TestCase


def pytest_runtest_setup(item):
    phasesTest = PhasesTest({})
    phasesTest.prepareConfig()
    testConfig = phasesTest.loadConfig("tests/config.yml")
    phasesTest.config.update(testConfig)
    project = phasesTest.createProjectFromConfig(phasesTest.config)
    TestCase.project = project




from .recordloaders.RecordLoaderAlice import RecordLoaderAlice
from .recordloaders.RecordLoaderDomino import RecordLoaderDomino

from .recordwriter.RecordWriterEDF import RecordWriterEDF
from .recordwriter.RecordWriterDICOM import RecordWriterDICOM



import numpy as np
from .RecordLoaderAlice import RecordLoaderAlice
from pyPhasesRecordloader import Event, RecordSignal, Signal


class RecordLoaderTest(RecordLoaderAlice):
    config = {}

    def getRecordList(self):
        return [f"record-{i}" for i in range(501)]

    def getMetadata(self, recordName):
        intId = int(recordName[7:])
        patientName = "group" if intId < 10 else recordName

        return {
            "patient": patientName,
            "dataCount": 100,
        }
    
    def exist(self, recordId):
        return True
    
    def existAnnotation(self, recordId):
        return recordId != "record-20"
    
    def getSignalHeaders(self, recordId):
        return [{
            "type": "test",
        }]

    def loadRecord(self, recordName):
        intId = int(recordName[7:])

        recordSignal = RecordSignal()
        recordSignal.addSignal(Signal("test1", np.arange(512) + intId - 10, frequency=1))
        recordSignal.addSignal(Signal("test2", np.arange(512) + intId - 9, frequency=1))
        events = [Event("lightOff", 10), Event("lightOn", 510), Event("R", 0, 512)]
        events.append(Event("arousal", int(10 + intId), 5))
        return recordSignal, events



from pathlib import Path
from unittest.mock import patch

import pandas as pd

from pyPhases.test import mockLogger
from pyPhases.test.Mocks import OverwriteConfig
from pyPhases.test.TestCase import TestCase

from SleepHarmonizer.phases.LoadData import LoadData


class TestLoadData(TestCase):
    phase = LoadData()

    def config(self):
        return {"dataversion": {"recordIds": None, "groupBy": None}}
    
    def setUp(self):
        super().setUp()
        dataIds = ["metadata-tmp", "allDBRecordIds"]
        for dId in dataIds:
            dId = self.project.getDataFromName(dId).getDataId()
            Path(f"data/{dId}").unlink(missing_ok=True)
        self.patcher_exists = patch("pathlib.Path.exists", return_value=True)
        self.mock_exists = self.patcher_exists.start()
        
        self.patcher_records = patch("pyPhasesRecordloader.RecordLoader.RecordLoader.getRecordList")
        self.mock_records = self.patcher_records.start()
        
        self.patcher_metadata = patch("pyPhasesRecordloaderSHHS.recordLoaders.RecordLoaderSHHS.RecordLoaderSHHS.getMetaData") 
        self.mock_metadata = self.patcher_metadata.start()    
        
        self.mock_metadataDict = {}
        def get_consistent_metadata(recordId):
            return self.mock_metadataDict.get(recordId, {})
    
        self.mock_metadata.side_effect = get_consistent_metadata

    def tearDown(self):
        self.patcher_exists.stop()
        self.patcher_records.stop() 
        self.patcher_metadata.stop()
        super().tearDown()


    def assert_metadata_equals(self, expected_data):
        metadata = self.project.getData("metadata")
        pd.testing.assert_frame_equal(
            metadata,
            pd.DataFrame(expected_data).set_index("recordId", drop=False),
            check_like=True  # Ignore column order during comparison
        )

    @OverwriteConfig({"dataversion": {"recordIds": ["id1", "id2", "id3"]}})
    def testMainFixedRecordIds(self):
        self.assertDataEqual("allDBRecordIds", {
            "id1": ["id1"], 
            "id2": ["id2"], 
            "id3": ["id3"]
        })

    def testMainRecordloader(self):
        self.mock_records.return_value = ["1", "2", "3"]
        self.mock_metadata.return_value = {}
        
        self.assertDataEqual("allDBRecordIds", {
            "1": ["1"], 
            "2": ["2"], 
            "3": ["3"]
        })

    def test_Metadata(self):
        self.mock_records.return_value = ["1", "2", "3"]
        self.mock_metadataDict = {
            "1": {"mydata": "A"},
            "2": {"mydata": "B"},
            "3": {"mydata": "C"}
        }

        expected = [
            {"recordId": "1", "metaDataExist": True, "annotationExist": True, "mydata": "A"},
            {"recordId": "2", "metaDataExist": True, "annotationExist": True, "mydata": "B"},
            {"recordId": "3", "metaDataExist": True, "annotationExist": True, "mydata": "C"}
        ]
        self.assert_metadata_equals(expected)

    def test_Metadata_Return(self):
        self.mock_records.side_effect = [["1", "2"], ["1", "2", "3"]]
        self.mock_metadataDict = {
            "1": {"mydata": "A"},
            "2": {"mydata": "B"},
            "3": {"mydata": "C"}
        }

        expected_initial = [
            {"recordId": "1", "annotationExist": True, "metaDataExist": True,  "mydata": "A"},
            {"recordId": "2", "annotationExist": True, "metaDataExist": True,  "mydata": "B"}
        ]
        self.assert_metadata_equals(expected_initial)

        self.project.unregister("metadata")
        self.project.unregister("metadata-tmp")

        expected_final = [
            {"recordId": "1", "annotationExist": True,"metaDataExist": True, "mydata": "A"},
            {"recordId": "2", "annotationExist": True,"metaDataExist": True, "mydata": "B"},
            {"recordId": "3", "annotationExist": True,"metaDataExist": True, "mydata": "C"}
        ]
        self.assert_metadata_equals(expected_final)

    def test_Metadata_specific(self):
        self.mock_records.side_effect = [["1", "2"], ["1", "2", "3"]]
        self.mock_metadataDict = {
            "1": {"mydata": "A"},
            "2": {"mydata": "B"},
            "3": {"mydata": "C"}
        }
        expected_record = {"recordId": "2", 'metaDataExist': True, "annotationExist": True, "mydata": "B"}

        x = self.project.getData("metadata", pd.DataFrame, recordId="2")
        self.assertEqual(len(x), 1, "error in generating specific record metadata")
        self.assertEqual(x.iloc[0].to_dict(), expected_record, "error in generating specific record metadata")

        x = self.project.getData("metadata", pd.DataFrame, recordId="2")
        self.assertEqual(len(x), 1, "error in loading specific record metadata")
        self.assertEqual(x.iloc[0].to_dict(), expected_record, "error in loading specific record metadata")

        # self.project.unregister("metadata")
        
        x = self.project.getData("metadata", pd.DataFrame, recordId="2")
        self.assertEqual(len(x), 1, "error in loading specific record metadata, during read")
        self.assertEqual(x.iloc[0].to_dict(), expected_record, "error in loading specific record metadata, during read")

    @OverwriteConfig({"dataversion": {"filterQuery": "score > 80"}})
    def test_filter_queries(self):
        self.mock_records.return_value = ["1", "2", "3"]
        self.mock_metadataDict = {
            "1": {"mydata": "A", "score": 90},
            "2": {"mydata": "B", "score": 85},
            "3": {"mydata": "C", "score": 70}
        }
        
        self.assertDataEqual("allDBRecordIds", {
            "1": ["1"],
            "2": ["2"]
        })

    @OverwriteConfig({"dataversion": {"channelFilterQuery": "type == 'eeg'"}})
    def test_channel_filter(self):
        self.mock_records.return_value = ["1", "2", "3"]
        self.mock_metadataDict = {
            "1": {"channels": [{"signalName": "EEG A1:C2", "type": "eeg"}]},
            "2": {"channels": [{"signalName": "EOG LLeft", "type": "eog"}]},
            "3": {"channels": [{"signalName": "EEG C3:A2", "type": "eeg"}]}
        }

        self.assertDataEqual("allDBRecordIds", {
            "1": ["1"],
            "3": ["3"]
        })

    @OverwriteConfig({"dataversion": {"filterQuery": "score > 80","channelFilterQuery": "type == 'eeg' and frequency >= 200"}})
    def test_combined_filters(self):
        self.mock_records.return_value = ["1", "2", "3", "4"]
        self.mock_metadataDict = {
            "1": {"score": 90, "channels": [{"signalName": "EEG A1:C2", "type": "eeg", "frequency": 250}]},
            "2": {"score": 85, "channels": [{"signalName": "EOG LLeft", "type": "eog", "frequency": 250}]},
            "3": {"score": 70, "channels": [{"signalName": "EEG C3:A2", "type": "eeg", "frequency": 250}]},
            "4": {"score": 70, "channels": [{"signalName": "EEG C3:A2", "type": "eeg", "frequency": 50}]}
        }

        self.assertDataEqual("allDBRecordIds", {
            "1": ["1"]
        })

    def test_MetadataDiffrentCols(self):
        self.mock_records.return_value = ["1", "2", "3"]
        self.mock_metadataDict = {
            "1": {"mydata": "A"},
            "2": {"mydata2": "B"},
            "3": {"mydata": "C"}
        }

        expected = [
            {"recordId": "1", 'metaDataExist': True, "annotationExist": True, "mydata": "A"},
            {"recordId": "2", 'metaDataExist': True, "annotationExist": True, "mydata2": "B"},
            {"recordId": "3", 'metaDataExist': True, "annotationExist": True, "mydata": "C"}
        ]
        self.assert_metadata_equals(expected)


from datetime import datetime
from typing import List

import pyedflib
from pyPhasesRecordloader import Event, Signal

from SleepHarmonizer.recordwriter.RecordWriter import RecordWriter


class RecordWriterEDF(RecordWriter):

    def getFilePath(self, recordName):
        return f"{self.filePath}/{recordName}.edf"

    def writeSignals(self, recordName, signals: List[Signal], events=None, startTime: datetime = None, signalIsDigital=False, force=False):

        if self.exist(recordName) and not force:
            self.log(f"Record {recordName} skipped because it already exists")
            return
        
        # write an edf file
        if events is None:
            events = []
 
        filePath = self.getFilePath(recordName)
        signalCount = len(signals)

        writer = pyedflib.EdfWriter(filePath, signalCount)

        # if self.patient is not None:
        #     writer.setPatientName(self.patient.name)
        #     writer.setGender(self.patient.gender)

        index = 0
        signalArray = []
        for index, signal in enumerate(signals):
            writer.setSignalHeader(
                index,
                {
                    "label": signal.name,
                    "dimension": signal.dimension,
                    "sample_frequency": signal.frequency,
                    "physical_max": signal.physicalMax,
                    "physical_min": signal.physicalMin,
                    "digital_min": signal.digitalMin,
                    "digital_max": signal.digitalMax,
                    "transducer": signal.transducer,
                    "prefilter": signal.prefilter,
                },
            )
            signalArray.append(signal.signal)
            index += 1
        if startTime is not None:
            writer.setStartdatetime(startTime)

        if signalIsDigital:
            signalArray = [s.astype("int32") for s in signalArray]

        writer.writeSamples(signalArray, digital=signalIsDigital)

        for annotation in events:
            annotation = annotation.todict() if isinstance(annotation, Event) else annotation
            duration = 0 if annotation["duration"] is None else annotation["duration"]
            writer.writeAnnotation(annotation["start"], duration, annotation["name"])

        writer.close()



import unittest

import numpy as np
from pyPhasesRecordloader import RecordSignal, Signal

from SleepHarmonizer.SignalPreprocessing import SignalPreprocessing


class TestSignalPreprocessing(unittest.TestCase):
    def setUp(self):
        self.preprocessor = SignalPreprocessing({"stepsPerType": {}})

        recordSignal = RecordSignal("test_record", 100)
        recordSignal.addSignal(Signal("ch1", np.random.randn(100), frequency=100))
        recordSignal.addSignal(Signal("ch2", np.random.randn(100), frequency=100))
        self.recordSignal = recordSignal

    def test_zerophase(self):
        b = np.array([0.1, 0.2, 0.3])
        a = np.array([1.0, -0.3, 0.2])
        x = np.array([1.0, 2.0, 3.0, 4.0])

        result = self.preprocessor._zerophase(b, a, x)

        # Add your assertions here
        # For example, to check type:
        self.assertIsInstance(result, np.ndarray)

        # To check against known values:
        expected_result = np.array([0.60434229, 0.9425683, 0.535641, 0.18467])  # Replace with the actual expected result
        np.testing.assert_array_almost_equal(result, expected_result, decimal=5)



from pathlib import Path

from SleepHarmonizer.PSGEventManager import PSGEventManager
from pyPhasesRecordloader.recordLoaders.EDFRecordLoader import EDFRecordLoader
from pyPhasesRecordloader import Event


from .AliceRMLLoader import AliceRMLLoader
from .AliceTextReportLoader import AliceTextReportLoader


class RecordLoaderAlice(EDFRecordLoader):
    def __init__(self, filePath, targetSignals, targetSignalTypes, optionalSignals=None, combineChannels=None) -> None:

        super().__init__(
            filePath=filePath,
            targetSignals=targetSignals,
            targetSignalTypes=targetSignalTypes,
            optionalSignals=optionalSignals,
            combineChannels=combineChannels,
        )
        self.validateOnly = False
        self.exportsEventArray = True

    def getFilePathSignal(self, recordId):
        return f"{self.filePath}/{recordId}/{recordId}.edf"

    def getFilePathAnnotation(self, recordId):
        return f"{self.filePath}/{recordId}/{recordId}.rml"

    def getFilePathTxt(self, recordId):
        return f"{self.filePath}/{recordId}/{recordId}.txt"

    def existAnnotation(self, recordId):
        return Path(self.getFilePathAnnotation(recordId)).exists()

    def exist(self, recordId):
        return self.existAnnotation() and Path(self.getFilePathSignal(recordId)).exists()

    def getEventList(self, recordName, targetFrequency=1):
        metaXML = self.getFilePathAnnotation(recordName)
        rmlLoader = self.getAliceLoader()
        eventArray = rmlLoader.loadAnnotation(metaXML)
        self.lightOff = rmlLoader.lightOff
        self.lightOn = rmlLoader.lightOn

        if self.lightOn is not None:
            eventArray.append(Event(name="lightOn", start=self.lightOn, frequency=0))

        if self.lightOff is not None:
            eventArray.append(Event(name="lightOff", start=self.lightOff, frequency=0))

        if targetFrequency != 1:
            [event.updateFrequency(targetFrequency) for event in eventArray]

        return self.fixEvents(eventArray)

    def getAliceLoader(self):

        rmlLoader = AliceRMLLoader()
        rmlLoader.classificationConfig = self.classificationConfig
        rmlLoader.channelMap = self.channelMap

        return rmlLoader

    # def fillRecordFromTxtReport(self, recordName, record):
    #     reportFile = self.getFilePath(recordName) + ".txt"
    #     loader = AliceTextReportLoader()
    #     loader.addTxtReport(reportFile, record)

    def getMetaData(self, recordName):

        metaData = super().getMetaData(recordName)
        metaData.update(self.getAliceLoader().getMetaData(self.getFilePathAnnotation(recordName)))
        metaData.update(AliceTextReportLoader().getMetaData(self.getFilePathTxt(recordName)))

        return metaData

    def fixArousal(self, df):
        em = PSGEventManager()

        em.dfAppendGroups(
            df,
            "arousal",
            "sleepStage",
            newGroupName="sleepStage-15",
            offsetStart=-15.1,
            fixedDuration=15.1,
        )

        removeQuery = "(name == 'arousal' and `sleepStage-15` == 'W')"
        rem = df.query(removeQuery)
        df.drop(rem.index, inplace=True)

        # if labelName == "SleepArousalsExtended":
        #     em.dfAppendGroups(
        #         df,
        #         "arousal",
        #         "limb",
        #         newGroupName="limb-arousal",
        #         offsetStart=-0.6,
        #         offsetEnd=0.6,
        #     )
        #     em.dfAppendGroups(
        #         df,
        #         "arousal",
        #         "sleepStage",
        #         newGroupName="woke-arousal",
        #         offsetStart=-0.5,
        #         offsetEnd=15.1,
        #     )
        #     df.loc[df.query("name == 'arousal' and `limb-arousal` != ''").index, "name"] = "arousal_limb"
        #     df.loc[df.query("name == 'arousal'").index, "name"] = "arousal_none"
        #     df.loc[df.query("name == 'arousal_rera'").index, "name"] = "arousal_rera_man"
        return df

    def fixSPO2Events(self, df):
        em = PSGEventManager()
        em.dfAppendGroups(
            df,
            "spo2",
            "sleepStage",
            newGroupName="sleepStage-15",
            offsetStart=-15.1,
            fixedDuration=15.2,
        )

        df["o2diff"] = df.apply(
            lambda row: (
                (float(row["data"]["O2Before"]) - float(row["data"]["O2Min"]))
                if "O2Min" in row["data"] and "O2Before" in row["data"]
                else None
            ),
            axis=1,
        )
        removeQuery = "(group == 'spo2' and manual == False and (o2diff < 3 or (`sleepStage-15` == 'W')))"
        rem = df.query(removeQuery, engine="python")

        df.drop(rem.index, inplace=True)
        return df
    
    def fixApnea(self, df):
        em = PSGEventManager()

        em.dfAppendGroups(
            df,
            "apnea",
            "sleepStage",
            newGroupName="sleepStageStart15",
            offsetStart=-15.1,
            fixedDuration=15.1,
        )

        arousalOffset = 5
        em.dfAppendGroups(
            df,
            "apnea",
            "arousal",
            newGroupName="arousal",
            offsetStart=0,
            offsetEnd=arousalOffset,
        )

        desatOffset = 10
        em.dfAppendGroups(
            df,
            "apnea",
            "spo2",
            newGroupName="desaturation",
            offsetStart=0,
            offsetEnd=desatOffset,
        )

        removeQuery = "(group == 'apnea' and manual == False  and (duration < 10 or sleepStageStart15 == 'W' or (name == 'resp_hypopnea' and arousal != 'arousal' and desaturation == '')))"
        rem = df.query(removeQuery)
        df.drop(rem.index, inplace=True)
        return df
    
    def fixLM(self, df):
        em = PSGEventManager()
        em.dfAppendGroups(
            df,
            "limb",
            "sleepStage",
            newGroupName="sleepStage-15",
            offsetStart=-15.1,
            fixedDuration=15.2,
        )
        em.dfAppendGroups(
            df,
            "limb",
            "apnea",
            newGroupName="apnea",
            offsetStart=-0.5,
            offsetEnd=0.5,
        )
        em.dfAppendGroups(
            df,
            "limb",
            "arousal",
            newGroupName="arousal",
            offsetStart=-0.5,
            offsetEnd=0.5,
        )

        removeQuery = (
            "(group == 'limb' and (`sleepStage-15` == 'W' or apnea != '' or arousal.str.contains('arousal_rera')))"
        )
        rem = df.query(removeQuery, engine="python")
        df.drop(rem.index, inplace=True)
        return df

    def fixEvents(self, eventArray):
        em = PSGEventManager()
        df = em.getDataframeFromEvents(eventArray)
        # legmovement depend on apnea + arousal
        # apnea depends on spo2events + arousal
        df = self.fixArousal(df)
        df = self.fixSPO2Events(df)
        df = self.fixApnea(df)
        df = self.fixLM(df)
        return [Event.fromdict(r) for r in df.T.to_dict().values()]



from datetime import datetime
from typing import List
import os
import numpy as np
import pydicom
from pydicom.dataset import Dataset
from SleepHarmonizer.PSGEventManager import PSGEventManager
from pyPhases.util.Logger import classLogger
from pyPhasesRecordloader import Event, Signal
from pydicom.fileset import FileSet

from SleepHarmonizer.recordwriter.RecordWriter import RecordWriter


@classLogger
class RecordWriterDICOM(RecordWriter):
    def _toDecimalString(self, number):
        return f"{float(number):.10g}"
    def getSubjectId(self):
        pc = self.metaData["patientCode"] if "patientCode" in self.metaData else "unknown"
        return f"sub-{pc}"

    def getSessionId(self):
        pc = self.metaData["sessionId"] if "sessionId" in self.metaData else "1"
        return f"ses-{pc}"

    def getFilePath(self, recordName):
        bids_path = os.path.join(self.filePath, self.getSubjectId(), self.getSessionId())

        return bids_path

    def getTaskFilePath(self, recordName, subject_id, session, sigType):
        basePath = self.getFilePath(recordName)
        file_name = f"{subject_id}_{session}_task-sleep_{sigType}.dcm"

        return f"{basePath}/{file_name}"

    def writeSignals(
        self, recordName, channels, events=None, startTime=None, signalIsDigital=False
    ):
        if events is None:
            events = []

        # Store study UID for SR reference
        self.study_instance_uid = pydicom.uid.generate_uid()

        # Split channels by modality
        channelsByType = {
            "eeg": [],
            "eog": [],
            "emg": [],
            "ecg": [],
            "sao2": [],
            "body": [],
            "effort": [],
            "flow": [],
            "mic": [],
            "resp": [],
        }

        channelMap = {
            "flow": "resp",
            "effort": "resp",
        }

        for ch in channels:
            if ch.typeStr in channelsByType:
                channel = channelMap[ch.typeStr] if ch.typeStr in channelMap else ch.typeStr
                channelsByType[channel].append(ch)

        # SOP Overview: https://dicom.nema.org/medical/dicom/current/output/chtml/part04/sect_b.5.html
        sopMap = {
            "eeg": "1.2.840.10008.5.1.4.1.1.9.7.4",
            "eog": "1.2.840.10008.5.1.4.1.1.9.7.3",
            "emg": "1.2.840.10008.5.1.4.1.1.9.7.2",
            "ecg": "1.2.840.10008.5.1.4.1.1.9.1.2",
            "body": "1.2.840.10008.5.1.4.1.1.9.8.1",
            # "effort": "1.2.840.10008.5.1.4.1.1.9.6.1",
            "resp": "1.2.840.10008.5.1.4.1.1.9.6.2",
            "mic": "1.2.840.10008.5.1.4.1.1.9.4.2",
            "sao2": "1.2.840.10008.5.1.4.1.1.9.5.1",
        }

        subject_id = self.getSubjectId()
        session = self.getSessionId()
        fs = FileSet()
        psg_em = PSGEventManager()
        groups = psg_em.getEventGroupMap()
        for sigType, sop in sopMap.items():
            if len(channelsByType[sigType]) == 0:
                continue

            eeg_dataset = self._create_dataset(
                sop_class_uid=sop,
                channels=channelsByType[sigType],
                events=events,
                modality=sigType,
            )
            
            match sigType:
                case "eeg" | "eog":
                    sig_groups = ["sleepStage", "arousal", "grapho"]
                case "emg":
                    sig_groups = ["limb"]
                case "ecg":
                    sig_groups = ["sleepStage", "arousal", "grapho", "limb", "apnea", "spo2"]
                    # sig_groups = ["cardiac"]
                case "effort":
                    sig_groups = ["apnea"]
                case "resp":
                    sig_groups = ["apnea", "spo2"]

            sig_groups.append("light")
            dic_events = [e for e in events if groups[e["name"]] in sig_groups]
            self._add_events_to_dataset(eeg_dataset, dic_events)

            # Get BIDS-compliant file paths (requires patient info)
            file_path = self.getTaskFilePath(recordName, subject_id, session, sigType)
            if not Path(file_path).parent.exists():
                Path(file_path).parent.mkdir(parents=True)
            eeg_dataset.save_as(file_path, write_like_original=False)
            fs.add(file_path)

        # Write SR if events exist
        if events:
            sr_path = self.writeAnnotationSR(recordName, events)
            fs.add(sr_path)

        fs.write(self.getFilePath(recordName) + ".dcm")
    
    def _create_basedataset(self, sop):
        ds = Dataset()
        ds.SOPClassUID = sop
        ds.StudyInstanceUID = self.study_instance_uid
        ds.SeriesInstanceUID = pydicom.uid.generate_uid()
        ds.SOPInstanceUID = pydicom.uid.generate_uid()
        ds.StudyDate = self.metaData["start"].strftime("%Y%m%d")
        ds.StudyTime = self.metaData["start"].strftime("%H%M%S.%f")
        ds.StudyID = "acq-1"
        ds.AccessionNumber = "123456789"
        ds.SeriesNumber = 1
        ds.ReferringPhysicianName = "Unknown"
        # ds.PerformedProcedureCodeSequence = self._create_code_sequence('LN', '28633-6', ' Polysomnography (sleep) study')

        # when Content was create (after manual scoring)
        ds.ContentDate = self.metaData["start"].strftime("%Y%m%d")
        ds.ContentTime = self.metaData["start"].strftime("%H%M%S.%f")
        ds.InstanceNumber = 1

        # Patient information
        ds.PatientName = self.metaData["patient"] if "patient" in self.metaData else self.metaData["patientName"]
        ds.PatientID = self.metaData["patientCode"]
        ds.PatientBirthDate = self.metaData["birthdate"]
        ds.PatientSex = self.metaData["sex"]

        # Equipment information
        for k, v in self.metaData["dicom"].items():
            setattr(ds, k, v)

        ds.is_little_endian = True
        ds.is_implicit_VR = True
        return ds

    def _create_dataset(self, sop_class_uid, channels, events, modality):
        ds = self._create_basedataset(sop_class_uid)
        
        ds.Modality = modality.upper()

        dateTime = self.metaData["start"]
        ds.AcquisitionDateTime = dateTime
        ds.AcquisitionContextSequence = []

        # Add waveform data
        self._add_channels_to_dataset(ds, channels)

        return ds

    # def _create_sidecar_json(self, dcm_path, channels, startTime):
    #     import json

    #     sidecar = {
    #         "TaskName": "sleep",
    #         "SamplingFrequency": channels[0].frequency,
    #         "RecordingDuration": len(channels[0].signal) / channels[0].frequency,
    #         "RecordingType": "continuous",
    #         "StartTime": startTime.isoformat() if startTime else None,
    #         "Manufacturer": "SleepHarmonizer",
    #         "PowerLineFrequency": 50,
    #         "SoftwareFilters": "n/a",
    #         "HardwareFilters": "n/a",
    #         "ChannelCount": len(channels),
    #         "Channels": {
    #             ch.name: {"sampling_frequency": ch.frequency, "units": ch.dimension, "type": ch.name.split("_")[0]}
    #             for ch in channels
    #         },
    #     }

    #     json_path = dcm_path.replace(".dcm", ".json")
    #     with open(json_path, "w") as f:
    #         json.dump(sidecar, f, indent=4)

    def _add_channels_to_dataset(self, dataset, channels):
        dataset.WaveformSequence = []

        for channel in channels:
            waveform_seq = Dataset()
            waveform_seq.MultiplexGroupLabel = channel.name
            waveform_seq.MultiplexGroupTimeOffset = 0
            waveform_seq.WaveformOriginality = "ORIGINAL"
            waveform_seq.NumberOfWaveformChannels = 1
            waveform_seq.NumberOfWaveformSamples = len(channel.signal)
            waveform_seq.SamplingFrequency = channel.frequency

            # Calculate bits needed
            diff = channel.digitalMax - channel.digitalMin
            required_bits = np.ceil(np.log2(diff))
            digital_center_diff = (channel.digitalMax - channel.digitalMin + 1) / 2 + channel.digitalMin

            # Set bit depth
            if required_bits <= 16:
                bits = 16
                sample_interp = "SS"
            else:
                bits = 32
                sample_interp = "SL"
            # Channel Source Sequence
            chan_src_seq = Dataset()
            chan_src_seq.CodeMeaning = channel.name

            # Channel definition
            chan_def = Dataset()
            chan_def.ChannelLabel = channel.name
            chan_def.WaveformBitsStored = bits
            chan_def.ChannelSourceSequence = [chan_src_seq]

            # Add sensitivity units
            unit_map = {
                "uV": ("μV", "microvolts"),
                "μV": ("μV", "microvolts"),
                "mV": ("mV", "millivolts"),
                "%": ("%", "percent"),
                "cmH2O": ("cm[H2O]", "centimeters of water"),
                "mmHg": ("mm[Hg]", "millimeters of mercury"),
                "BPM": ("/min", "beats per minute"),
                "bpm": ("/min", "beats per minute"),
                "Hz": ("Hz", "hertz"),
                "°C": ("Cel", "degrees Celsius"),
                "celsius": ("Cel", "degrees Celsius"),
                "SpO2": ("%", "oxygen saturation percentage"),
                "L/min": ("L/min", "liters per minute"),
                "mL/min": ("mL/min", "milliliters per minute"),
                "V": ("V", "volts"),
                "A": ("A", "amperes"),
                "Ohm": ("Ohm", "ohms"),
                "g": ("g", "grams"),
                "kg": ("kg", "kilograms")
            }
            
            if channel.dimension in unit_map:
                ucum_code, description = unit_map[channel.dimension]
                chan_def.ChannelSensitivityUnitsSequence = self._create_code_sequence("UCUM", ucum_code, description)

            # Calculate scaling factors
            digital_to_physical = (channel.physicalMax - channel.physicalMin) / (channel.digitalMax - channel.digitalMin)
            baseline = (channel.physicalMax + channel.physicalMin) / 2 + digital_center_diff * digital_to_physical

            # chan_def.ChannelSensitivity = digital_to_physical
            chan_def.ChannelSensitivity =  self._toDecimalString(digital_to_physical)

            chan_def.ChannelBaseline = baseline
            chan_def.ChannelSampleSkew = 0
            chan_def.ChannelSensitivityCorrectionFactor = 1

            waveform_seq.ChannelDefinitionSequence = [chan_def]
            waveform_seq.WaveformBitsAllocated = bits
            waveform_seq.WaveformSampleInterpretation = sample_interp
            waveform_seq.WaveformData = (channel.signal - digital_center_diff).astype(f"int{bits}").reshape(-1, 1).tobytes()

            dataset.WaveformSequence.append(waveform_seq)

    def _add_events_to_dataset(self, dataset, events, mainCodeSeq=None):
        mainCodeSeq = ("DSM", "130860", "Pattern Event") if mainCodeSeq is None else mainCodeSeq
        dataset.WaveformAnnotationSequence = []
        for event in events:
            annotation = Dataset()
            annotation.ReferencedWaveformChannels = [0]
            if event["duration"] == 0:
                annotation.TemporalRangeType = "POINT"
                annotation.ReferencedTimeOffsets = [self._toDecimalString(event["start"])]
            else:
                annotation.TemporalRangeType = "SEGMENT"
                annotation.ReferencedTimeOffsets = [
                    self._toDecimalString(event["start"]), 
                    self._toDecimalString(event["end"])
                ]
            # annotation.ReferencedSamplePositions = [
            #     int(event["start"]), 
            #     int(event["start"] + event["duration"])
            # ]
            # annotation.ReferencedSamplePositions = [
            #     int(event["start"])
            # ]
            # annotation.TemporalRangeType = "POINT"
            # annotation.ReferencedTimeOffsets = [self._toDecimalString(event["start"])]
            # annotation.UnformattedTextValue = event["name"]
            codeSeq = self.getEventCodeSequence(event["name"])
            if codeSeq is None:
                name = event["name"]
                self.logError(f"Codesequence not found for event '{name}'")
                continue
            annotation.ConceptNameCodeSequence = self._create_code_sequence(*codeSeq)
            annotation.ConceptCodeSequence = self._create_code_sequence(*codeSeq)
            # annotation.ConceptCodeSequence = self._create_code_sequence("MDC", "10:256", "P wave")
            dataset.WaveformAnnotationSequence.append(annotation)

    def writeAnnotationSR(self, recordName, events):
        # Create SR dataset
        sr = self._create_basedataset('1.2.840.10008.5.1.4.1.1.88.22')
        
        sr.Modality = 'SR'      
        sr.CompletionFlag = "PARTIAL"
        sr.VerificationFlag = "UNVERIFIED"
        sr.ValueType = "CODE"

        sr.PerformedProcedureCodeSequence = self._create_code_sequence('LN', '28633-6', ' Polysomnography (sleep) study')
        
        sr.ConceptNameCodeSequence = self._create_code_sequence('LN', '28633-6', ' Polysomnography (sleep) study')
        sr.ConceptCodeSequence = self._create_code_sequence('DCM', '130868', ' Neurophysiology Post-hoc Review Annotations')

        # DCM 130868 Neurophysiology Post-hoc Review Annotations


        ref_proc_step = Dataset()
        ref_proc_step.ReferencedSOPClassUID = '1.2.840.10008.5.1.4.1.1.9.7.4'
        ref_proc_step.ReferencedSOPInstanceUID = uid.generate_uid()
        sr.ReferencedPerformedProcedureStepSequence = [ref_proc_step]


        sr.ContentSequence = self.getEvents(events)

        # Save SR document
        sr_path = os.path.join(self.getFilePath(recordName), f"{self.getSubjectId()}_{self.getSessionId()}_annotations.sr.dcm")
        sr.save_as(sr_path, write_like_original=False)
        return sr_path

    def _create_code_sequence(self, scheme_designator, code_value, code_meaning):
        seq = Dataset()
        seq.CodeValue = code_value
        seq.CodingSchemeDesignator = scheme_designator 
        seq.CodeMeaning = code_meaning
        return [seq]
    
    def getEvents(self, events):
        
        items = [self.create_event(e["name"], e["start"], e["duration"], i) for i, e in enumerate(events)]
        items = [item for item in items if item is not None]

        return items
    
    def getEventCodeSequence(self, name):
        cid_mapping = {
            'sleepStage': {
                'W': ('MDC', '2:23672', 'Sleep stage wake'),
                'N1': ('DCM', '130834', 'Sleep stage N1'), 
                'N2': ('DCM', '130835', 'Sleep stage N2'), 
                'N3': ('DCM', '130836', 'Sleep stage N3'), 
                # 'N1': ('MDC', '2:23696', 'Sleep stage N1'), 
                # 'N2': ('MDC', '2:23704', 'Sleep stage N2'),
                # 'N3': ('MDC', '2:23712', 'Sleep stage N3'),
                'R': ('MDC', '2:23680', 'Sleep stage REM')
            },
            'apnea': {
                'resp_obstructiveapnea': ('MDC', '3:3072', 'Apnea (obstructive)'),
                'resp_centralapnea': ('MDC', '3:3072', 'Apnea (central)'),
                'resp_mixedapnea': ('MDC', '3:3072', 'Apnea (mixed)'),
                'resp_hypopnea': ('MDC', '3:3072', 'Hypopnea')
            },
            'arousal': {
                'arousal': ('MDC', '2:23800', ' Sleep arousal'),
                'arousal_rera':  ('MDC', '2:23800', ' Sleep arousal (RERA)'),
                'arousal_plm': ('MDC', '2:23800', ' Sleep arousal (Periodic limb movement)'),
                'arousal_spontaneous': ('MDC', '2:23800', ' Sleep arousal (Spontaneous)'),
                'arousal_respiratory': ('MDC', '2:23800', ' Sleep arousal (Respiratory event)'),
            },
            'limb': {
                'LegMovement-Left': ('MDC', '2:24184', 'Periodic movements of sleep'),
                'LegMovement-Right': ('MDC', '2:24184', 'Periodic movements of sleep'),
                'LegMovement': ('MDC', '2:24184', 'Periodic movements of sleep'),
                #  Periodic movements of sleep with arousals
            },
            'light': {
                'lightOff': ('MDC', 'MDC_EVT_STAT_LIGHTS_IN_ROOM_OFF', 'Lights off'),
                'lightOn': ('MDC', 'MDC_EVT_STAT_LIGHTS_IN_ROOM_ON', 'Lights on'),
            }
        }
        psg_em = PSGEventManager()
        groups = psg_em.getEventGroupMap()
        event_type = groups[name]
        if event_type in cid_mapping and name in cid_mapping[event_type]:
            return cid_mapping[event_type][name]
        
        return None


    def create_event(self, name, start_time, duration, event_number):
        
        code_sequence = self.getEventCodeSequence(name)
        
        if code_sequence is not None:
            event_container = Dataset()
            event_container.ValueType = "CONTAINER"
            event_container.RelationshipType = "CONTAINS"
            
            event_container.ConceptNameCodeSequence = self._create_code_sequence(f"11110{event_number}", "DCM", f"Sleep Apnea Episode {event_number}")
            event_container.ConceptCodeSequence = self._create_code_sequence(f"11110{event_number}", "DCM", f"Sleep Apnea Episode {event_number}")
            event_container.ContentSequence = []

            # Event Type
            event = Dataset()
            event.ValueType = "CODE"
            event.ConceptNameCodeSequence = self._create_code_sequence(code_sequence[0], code_sequence[1], code_sequence[2])
            # event.ConceptCodeSequence = _create_code_sequence(code_sequence[0], code_sequence[1], code_sequence[2])
            event_container.ContentSequence.append(event)

            # Start Time
            start_time_ds = Dataset()
            start_time_ds.ValueType = "NUM"
            start_time_ds.ConceptNameCodeSequence = self._create_code_sequence("111400", "DCM", "Start of Event")
            start_time_ds.MeasuredValueSequence = [Dataset()]
            start_time_ds.MeasuredValueSequence[0].NumericValue = start_time
            start_time_ds.MeasuredValueSequence[0].MeasurementUnitsCodeSequence = [Dataset()]
            start_time_ds.MeasuredValueSequence[0].MeasurementUnitsCodeSequence = self._create_code_sequence("s", "UCUM", "Seconds")
            event_container.ContentSequence.append(start_time_ds)

            # Duration
            duration_ds = Dataset()
            # duration_ds.RelationshipType = "HAS_OBS_CONTEXT"
            duration_ds.ValueType = "NUM"
            duration_ds.ConceptNameCodeSequence = [Dataset()]
            duration_ds.ConceptNameCodeSequence = self._create_code_sequence( "DCM", "111401", "Duration")

            duration_ds.MeasuredValueSequence = [Dataset()]
            duration_ds.MeasuredValueSequence[0].NumericValue = duration
            duration_ds.MeasuredValueSequence[0].MeasurementUnitsCodeSequence = self._create_code_sequence("UCUM", "s", "Seconds")
            event_container.ContentSequence.append(duration_ds)

            # # Referenced Waveform
            # waveform_ref = Dataset()
            # waveform_ref.ValueType = "IMAGE"
            # waveform_ref.ConceptNameCodeSequence = [Dataset()]
            # waveform_ref.ConceptNameCodeSequence[0].CodeValue = "121020"
            # waveform_ref.ConceptNameCodeSequence[0].CodingSchemeDesignator = "DCM"
            # waveform_ref.ConceptNameCodeSequence[0].CodeMeaning = "Referenced Waveform"
            # waveform_ref.ReferencedSOPSequence = [Dataset()]
            # waveform_ref.ReferencedSOPSequence[0].ReferencedSOPClassUID = "1.2.840.10008.5.1.4.1.1.9.1"
            # waveform_ref.ReferencedSOPSequence[0].ReferencedSOPInstanceUID = sop_uid
            # event_container.ContentSequence.append(waveform_ref)
        else:
            return None
        
        return event_container


from pyPhases import Data, PluginAdapter, Project
from pyPhasesRecordloader import RecordLoader

from SleepHarmonizer.recordwriter.RecordWriter import RecordWriter

from .phases.Export import Export
from .phases.LoadData import LoadData


class Plugin(PluginAdapter):
    def __init__(self, project: Project, options=None):
        super().__init__(project, options)

        if "LoadData" not in self.project.phaseMap:
            dataDep = [
                Data("metadata", self.project, ["dataBase"]),
                Data("metadata-channels", self.project, ["dataBase"]),
                Data("allDBRecordIds", self.project, ["metadata", "dataversion.recordIds", "dataversion.groupBy"]),
                Data("metadata-tmp", self.project, ["metadata"]),
                Data("metadata-channels-tmp", self.project, ["metadata-channels"]),
            ]
            loadData = LoadData(dataDep)
            project.addPhase(loadData)
        if "Export" not in self.project.phaseMap:
            project.addPhase(Export([]))

    def initPlugin(self):
        RecordLoader.registerRecordLoader("RecordLoaderTest", "SleepHarmonizer.recordloaders")
        RecordLoader.registerRecordLoader("RecordLoaderDomino", "SleepHarmonizer.recordloaders")
        RecordLoader.registerRecordLoader("RecordLoaderAlice", "SleepHarmonizer.recordloaders")

        RecordWriter.recordWriter.set(
            name=self.getConfig("useWriter"),
            options={"filePath": self.getConfig("export-path")},
            dynOptions={},
        )



from pathlib import Path
from SleepHarmonizer.recordloaders.DominoErgLoader import DominoErgLoader

from pyPhasesRecordloader.recordLoaders.EDFRecordLoader import EDFRecordLoader
from pyPhasesRecordloader import AnnotationNotFound
from .DominoAnnotationLoader import DominoAnnotationLoader


class RecordLoaderDomino(EDFRecordLoader):
    possibleNamesSleepstages = ["SchlafProfil", "Schlafprofil"]
    possibleRespEvents = ["Flow Events"]
    possibleArousalEvents = ["Klassifizierte Arousal"]

    def getFileBasePath(self, recrdId):
        return self.filePath + "/" + recrdId

    def getFilePathSignal(self, recordId):
        return self.getFileBasePath(recordId) + "/" + recordId + ".edf"

    def getFilePathAnnotation(self, recordId):
        annotationFolder = self.downloaderConfig.get("annotationFolder", "annotations")
        return self.getFileBasePath(recordId) + f"/{annotationFolder}"
    
    def getErgPath(self, recordId):
        return self.getFileBasePath(recordId) + "/" + recordId + ".erg"

    def existAnnotation(self, recordId):
        return [Path(self.getFilePathAnnotation(recordId) + f"/{n}.txt").exists() for n in self.possibleNamesSleepstages]

    def loadAnnotation(self, recordId, fileNames, valueMap=None, possibleValues=None):
        found = False
        for fileName in fileNames:
            filePath = self.getFilePathAnnotation(recordId)  + "/" + fileName + ".txt"
            if Path(filePath).exists():
               found = True
               break
        
        if not found:
            raise AnnotationNotFound(fileName)
            
        annotationLoader = DominoAnnotationLoader.load(filePath, valueMap, self.annotationFrequency, possibleValues)

        return annotationLoader.events


    def getMetaData(self, recordName):

        metaData = super().getMetaData(recordName)
        # metaData.update(self.getAliceLoader().getMetaData(self.getFilePathAnnotation(recordName)))
        metaData.update(DominoErgLoader().getMetaDataFromFile(self.getErgPath(recordName), DominoErgLoader.relevantRows))
        
        # lightOff = self.getXMLPath(self.metaXML, ["Acquisition", "Sessions", "Session", "LightsOff"])
        # off, on = self.getLightAnnotations()
        # metaData["lightOff"] = off
        # metaData["lightOn"] = on

        return metaData
    
    def getLightAnnotations(self):
        """get light off, on in seconds"""
        lightOff = self.getLastAnnotationTimeByName("Licht aus")
        lightOn = self.getFirstAnnotationTimeByName("Ende der Messung")
        
        if lightOff is None:
            self.logError("Die Annotation 'Licht Aus' wurde nicht gefunden")
        else:
            lightOff = int(lightOff)
        if lightOn is None:
            self.logWarning("Die Annotation 'Ende der Messung' wurde nicht gefunden")
        else:
            lightOn = int(lightOn)

        return lightOff, lightOn
            

    def getEventList(self, recordName, targetFrequency=1):
        self.annotationFrequency = targetFrequency

        eventArray = self.loadAnnotation(
            recordName,
            self.possibleNamesSleepstages,
            {
                "N3": "N3",
                "N2": "N2",
                "N1": "N1",
                "Rem": "R",
                "Wach": "W",
                "Artefakt": "undefined",
            },
        )

        # eventArray = self.loadAnnotation(
        #     recordName,
        #     "Autonome Arousal",
        #     {
        #         "Autonome Arousal": "arousal",
        #     },
        # )
        try:
            eventArray += self.loadAnnotation(
                recordName,
                self.possibleRespEvents,
                {
                    "Obstruktive Apnoe": "resp_obstructiveapnea",
                    "Gemischte Apnoe": "resp_mixedapnea",
                    "Zentrale Apnoe": "resp_centralapnea",
                    "Hypopnoe": "resp_hypopnea",
                    "RERA": "arousal_rera",
                },
                possibleValues=["Obstruktive Apnoe", "Gemischte Apnoe", "Zentrale Apnoe", "Hypopnoe", "RERA"],
            )
        except AnnotationNotFound:
            pass

        try:
            eventArray += self.loadAnnotation(
                recordName,
                self.possibleArousalEvents,
                {

                    "Arousal": "arousal",
                    "Respiratorische Arousal": "arousal_rera",
                    "PLM Arousal": "arousal_plm",
                    "LM Arousal": "arousal_limb",
                    "Schnarchen Arousal": "arousal_snore",
                    "SpO2 Arousal": "arousal",

                    "Arousal (EEG)": "arousal",
                    "Respiratorische Arousal (EEG)": "arousal_rera",
                    "PLM Arousal (EEG)": "arousal_plm",
                    "LM Arousal (EEG)": "arousal_limb",
                    "Schnarchen Arousal (EEG)": "arousal_snore",                
                    "SpO2 Arousal (EEG)": "arousal",

                    "Arousal (Autonome)": "arousal",
                    "Respiratorische Arousal (Autonome)": "arousal_rera",
                    "PLM Arousal (Autonome)": "arousal_plm",
                    "LM Arousal (Autonome)": "arousal_limb",
                    "Schnarchen Arousal (Autonome)": "arousal_snore",                
                    "SpO2 Arousal (Autonome)": "arousal",
                },
                possibleValues=[
                    "Arousal", "Respiratorische Arousal", "PLM Arousal", "LM Arousal", "Schnarchen Arousal", "SpO2 Arousal",
                    "Arousal (EEG)", "Respiratorische Arousal (EEG)", "PLM Arousal (EEG)", "LM Arousal (EEG)", "Schnarchen Arousal (EEG)", "SpO2 Arousal (EEG)",
                    "Arousal (Autonome)", "Respiratorische Arousal (Autonome)", "PLM Arousal (Autonome)", "LM Arousal (Autonome)", "Schnarchen Arousal (Autonome)", "SpO2 Arousal (Autonome)",
                ],
            )
        except AnnotationNotFound:
            pass

        self.lightOff, self.lightOn = self.getLightAnnotations()

        return eventArray




from abc import ABC, abstractmethod
import datetime
from pathlib import Path
from typing import List
from pyPhases.util.Logger import classLogger
from pyPhasesRecordloader.Event import Event
from pyPhasesRecordloader.RecordSignal import Signal
from pyPhasesRecordloader.util.DynamicModule import DynamicModule

import SleepHarmonizer.recordwriter as recordManagerPath


@classLogger
class RecordWriter(ABC):
    recordWriter = DynamicModule(recordManagerPath)

    recordWriters = {
        "RecordWriterEDF": "SleepHarmonizer.recordwriter",
        "RecordWriterDICOM": "SleepHarmonizer.recordwriter",
    }

    @classmethod
    def get(cls) -> "RecordWriter":
        packageName = cls.recordWriters[cls.recordWriter.moduleName]
        return cls.recordWriter.get(packageName)

    def __init__(self, filePath = ".") -> None:
        self.filePath = filePath
        self.metaData = {}

        
    def getFilePath(self, recordName):
        return f"{self.filePath}/{recordName}"

    
    def createFolderStructure(self):
        p = Path(self.filePath)
        p.parent.mkdir(parents=True, exist_ok=True)

    def exist(self, recordName):
        return Path(self.getFilePath(recordName)).exists()
    
    @abstractmethod
    def writeSignals(self, recordName, channels: List[Signal], events: List[Event]=None, startTime: datetime = None, signalIsDigital=False, force=False):
        pass
        



from unittest.mock import patch

import numpy as np
from pyPhases.test.Mocks import OverwriteConfig
from pyPhases.test.TestCase import TestCase
from pyPhasesRecordloader import RecordSignal, Event, Signal

from SleepHarmonizer.phases.Export import Export


def getTestSignalData(i):
    record = RecordSignal(recordId="myId")
    testSignal = Signal(name="testsignal", signal=np.array([1, 2, 3]), frequency=1)
    record.addSignal(signal=testSignal)

    testSignal = Signal(name="testsignalNotTarget", signal=np.array([1, 2, 3]), frequency=1)
    record.addSignal(signal=testSignal)

    events = [Event(start=1, duration=1, name="arousal"), Event(start=1, duration=1, name="arousal_notlisted")]

    return record, events

def getSubjectId(i):
    return i


class TestLoadData(TestCase):
    phase = Export()

    @patch("shutil.move", return_value=None)
    @patch("pyPhasesRecordloaderSHHS.recordLoaders.RecordLoaderSHHS.RecordLoaderSHHS.getSubjectId", side_effect=lambda r: r)
    @patch("pyPhasesRecordloaderSHHS.recordLoaders.RecordLoaderSHHS.RecordLoaderSHHS.getMetaData", return_value={"mydata": "A"})
    @patch("pyPhasesRecordloaderSHHS.recordLoaders.RecordLoaderSHHS.RecordLoaderSHHS.loadRecord", side_effect=getTestSignalData)
    @patch("SleepHarmonizer.recordwriter.RecordWriterEDF.RecordWriterEDF.writeSignals", return_value=None)
    @OverwriteConfig({"recordId": "myId"})
    def testExport(self, mock_writeSignals, mock_loadRecord, metadata_mock, mock_subjectid, mock_move):
        self.phase.run()

        self.assertTrue(mock_writeSignals.called)
        self.assertTrue(mock_loadRecord.called)

        file = mock_writeSignals.call_args[0][0]
        signals = mock_writeSignals.call_args[0][1]
        recordEvents = list(mock_writeSignals.call_args[0][2])

        self.assertEqual(file, "myId-tmp")
        self.assertEqual(len(signals), 1)
        self.assertEqual(len(recordEvents), 1)
        self.assertEqual(signals[0].name, "testsignal")
        self.assertEqual(recordEvents[0]["name"], "arousal")
        self.assertEqual(recordEvents[0]["start"], 1)
        self.assertEqual(recordEvents[0]["duration"], 1)
        self.assertEqual(signals[0].signal.tolist(), [1, 2, 3])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="SleepHarmonizer",
    version="v0.5.0"[1:],
    author="Franz Ehrlich",
    author_email="fehrlichd@gmail.com",
    description="A Plugin and stand alone tool to harmonize sleep related data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/sleep-is-all-you-need/sleep-harmonizer",
    packages=setuptools.find_packages(),
    package_data={"": ["*.yaml"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "pyPhases",
        "pyPhasesRecordloader",
        "pyPhasesRecordloaderPhysionet",
        "pyPhasesRecordloaderSHHS",
        "pyPhasesRecordLoaderMrOS",
        "pyPhasesRecordloaderMESA",
        "phases",
        "numpy",
        "tqdm",
        "pydicom",
        "pyedflib",
    ],
    python_requires=">=3.5",
)



from unittest import TestCase

from pyPhases import Project

from pyPhasesRecordloader import RecordLoader
from SleepHarmonizer.Plugin import Plugin


class TestPlugin(TestCase):
    def setUp(self):
        self.options = {}
        self.project = Project()
        self.project.config = self.project.loadConfig("SleepHarmonizer/config.yaml")
        self.plugin = Plugin(self.project, self.options)

    def test_project_is_extended(self):
        self.assertIn("Export", self.project.phaseMap)
        self.assertIn("LoadData", self.project.phaseMap)

        loadDataPhase = self.project.getPhase("LoadData")
        exportData = [d.name for d in loadDataPhase.exportData]
        self.assertIn("metadata", exportData)
        self.assertTrue("allDBRecordIds", exportData)

    def test_initPlugin(self):
        self.plugin.initPlugin()
        self.assertIn("RecordLoaderDomino", RecordLoader.recordLoaders.keys())
        self.assertIn("RecordLoaderAlice", RecordLoader.recordLoaders.keys())
